/*
 * debug - debug_io.c
 * 	version 0.0.1
 * 
 * Copyright (C) 2005, 2006, 2007  Krzysztof Mazur <krzysiek@podlesie.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * $Id: debug_io.c,v 1.9 2007/02/24 14:04:54 krzysiek Exp $
 */

#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <stdio.h>
#include <debug_io.h>
#include <inttypes.h>
#include <errno.h>

#ifdef HAVE_SYS_IO_H
#include <sys/io.h>
#define IOPERM_IO 
#elif HAVE_X86_PIO_H
#include <x86/pio.h>
#define IOPERM_IO 
#endif

#define IO_DEV_PORT

#ifdef IOPERM_IO
#ifdef HAVE_IOPERM
static int my_ioperm(unsigned long from, unsigned long num, int turn_on)
{
	return ioperm(from, num, turn_on) ? -errno : 0;
}
#else
static int my_ioperm(unsigned long from, unsigned long num, int turn_on)
{
	return -ENOSYS;
}
#endif

#ifdef HAVE_IOPL
static int my_iopl(int level)
{
	return iopl(level) ? -errno : 0;
}
#elif HAVE_MACHINE_SYSARCH_H
#include <machine/sysarch.h>
static int my_iopl(int level)
{
	return i386_iopl(level) ? -errno : 0;
}
#else
static int my_iopl(int level)
{
	return -ENOSYS;
}
#endif

static int iopl_counter;

int d_io_enable(unsigned long port, int num)
{
	int ret;
	
	ret = my_ioperm(port, num, 1);
	if (!ret)
		goto out;
	ret = iopl_counter ? 0 : my_iopl(3);
	iopl_counter += !ret;
out:
	return ret;
}

/*
 * d_io_disable() should be called with the same port and num arguments
 * as d_io_enable().
 */
int d_io_disable(unsigned long port, int num)
{
	int ret;
	
	ret = my_ioperm(port, num, 0);
	if (ret)
		ret = --iopl_counter ? 0 : my_iopl(0);
	return ret;
}

static int d_inb_ioperm(unsigned long port, uint8_t *value)
{
	int ret;

	if ((ret = d_io_enable(port, 1)))
		return ret;
	*value = inb(port);
	if ((ret = d_io_disable(port, 1)))
		return ret;
	return 0;
}

static int d_outb_ioperm(unsigned long port, uint8_t value)
{
	int ret;
	
	if ((ret = d_io_enable(port, 1)))
		return ret;
	outb(value, port);
	if ((ret = d_io_disable(port, 1)))
		return ret;
	return 0;
}

static int d_inw_ioperm(unsigned long port, uint16_t *value)
{
	int ret;
	
	if ((ret = d_io_enable(port, 2)))
		return ret;
	*value = inw(port);
	if ((ret = d_io_disable(port, 2)))
		return ret;
	return 0;
}

static int d_outw_ioperm(unsigned long port, uint16_t value)
{
	int ret;
	
	if ((ret = d_io_enable(port, 2)))
		return ret;
	outw(value, port);
	if ((ret = d_io_disable(port, 2)))
		return ret;
	return 0;
}

static int d_inl_ioperm(unsigned long port, uint32_t *value)
{
	int ret;
	
	if ((ret = d_io_enable(port, 4)))
		return ret;
	*value = inl(port);
	if ((ret = d_io_disable(port, 4)))
		return ret;
	return 0;
}

static int d_outl_ioperm(unsigned long port, uint32_t value)
{
	int ret;
	
	if ((ret = d_io_enable(port, 4)))
		return ret;
	outl(value, port);
	if ((ret = d_io_disable(port, 4)))
		return ret;
	return 0;
}

static int init_ioperm(void)
{
	d_inb = d_inb_ioperm;
	d_inw = d_inw_ioperm;
	d_inl = d_inl_ioperm;
	d_outb = d_outb_ioperm;
	d_outw = d_outw_ioperm;
	d_outl = d_outl_ioperm;
	return 0;
}

#else

int d_io_enable(unsigned long port, int num)
{
	return -ENOSYS;
}

int d_io_disable(unsigned long port, int num)
{
	return -ENOSYS;
}

#endif

static int d_inb_dummy(unsigned long port, uint8_t *value)
{
	return -ENOSYS;
}

static int d_inw_dummy(unsigned long port, uint16_t *value)
{
	return -ENOSYS;
}

static int d_inl_dummy(unsigned long port, uint32_t *value)
{
	return -ENOSYS;
}

static int d_outb_dummy(unsigned long port, uint8_t value)
{
	return -ENOSYS;
}

static int d_outw_dummy(unsigned long port, uint16_t value)
{
	return -ENOSYS;
}

static int d_outl_dummy(unsigned long port, uint32_t value)
{
	return -ENOSYS;
}

static int init_dummy(void)
{
	d_inb = d_inb_dummy;
	d_inw = d_inw_dummy;
	d_inl = d_inl_dummy;
	d_outb = d_outb_dummy;
	d_outw = d_outw_dummy;
	d_outl = d_outl_dummy;
	return 0;
}

#ifdef IO_DEV_PORT
static int io_dev_port_fd = -1;

static int d_in_port(unsigned long port, void *buf, size_t size)
{
	int ret;
	
	if (io_dev_port_fd == -1)
		return -EBADF;
	if (lseek(io_dev_port_fd, port, SEEK_SET) == -1)
		return -errno;
	ret = read(io_dev_port_fd, buf, size);
	return (ret >= 0) ? 0 : -errno;
}

static int d_out_port(unsigned long port, const void *buf, size_t size)
{
	int ret;
	
	if (io_dev_port_fd == -1)
		return -EBADF;
	if (lseek(io_dev_port_fd, port, SEEK_SET) == -1)
		return -errno;
	ret = write(io_dev_port_fd, buf, size);
	return (ret >= 0) ? 0 : -errno;
}

static int d_inb_port(unsigned long port, uint8_t *value)
{
	return d_in_port(port, value, 1);
}

static int d_inw_port(unsigned long port, uint16_t *value)
{
	return d_in_port(port, value, 2);
}

static int d_inl_port(unsigned long port, uint32_t *value)
{
	return d_in_port(port, value, 4);
}

static int d_outb_port(unsigned long port, uint8_t value)
{
	return d_out_port(port, &value, 1);
}

static int d_outw_port(unsigned long port, uint16_t value)
{
	return d_out_port(port, &value, 2);
}

static int d_outl_port(unsigned long port, uint32_t value)
{
	return d_out_port(port, &value, 4);
}

static int init_dev_port(void)
{
	io_dev_port_fd = open("/dev/port", O_RDWR);
	if (io_dev_port_fd == -1) {
		fprintf(stderr, "open /dev/port error: %s\n", strerror(errno));
		return init_dummy();
	}
	d_inb = d_inb_port;
	d_inw = d_inw_port;
	d_inl = d_inl_port;
	d_outb = d_outb_port;
	d_outw = d_outw_port;
	d_outl = d_outl_port;
	return 0;
}
#endif

int d_ioinit(int io_access_mode)
{
	if (io_access_mode == IOA_DEFAULT) {
#ifdef IOPERM_IO
		io_access_mode = IOA_IOPERM;
#elif defined(IO_DEV_PORT)
		io_access_mode = IOA_DEV_PORT;
#else
		io_access_mode = IOA_DUMMY;
#endif
	}

	switch (io_access_mode) {
#ifdef IOPERM_IO
		case IOA_IOPERM:
			init_ioperm();
			break;
#endif
#ifdef IO_DEV_PORT
		case IOA_DEV_PORT:
			init_dev_port();
			break;
#endif
		case IOA_DUMMY:
			init_dummy();
			break;
		default:
			fprintf(stderr, "Unkown I/O access mode. Disabling i/o "
					"support.\n");
			init_dummy();
			break;
	}
	return 0;
}
